from __future__ import annotations

import argparse


def _all_option_strings(parser: argparse.ArgumentParser) -> list[str]:
    options: list[str] = []
    for action in parser._actions:
        for option in action.option_strings:
            options.append(option)
    return sorted(set(options))


def _bash_script(parser: argparse.ArgumentParser) -> str:
    options = " ".join(_all_option_strings(parser))
    prog = parser.prog
    return f"""_{prog}_completion() {{
  local cur
  cur="${{COMP_WORDS[COMP_CWORD]}}"
  COMPREPLY=($(compgen -W "{options}" -- "$cur"))
}}
complete -F _{prog}_completion {prog}
"""


def _zsh_script(parser: argparse.ArgumentParser) -> str:
    options = " ".join(_all_option_strings(parser))
    prog = parser.prog
    return f"""#compdef {prog}
local -a opts
opts=({options})
_describe 'option' opts
"""


def _tcsh_script(parser: argparse.ArgumentParser) -> str:
    options = " ".join(_all_option_strings(parser))
    prog = parser.prog
    return f"complete {prog} 'p/*/({options})/'\n"


class _PrintCompletion(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        generators = {
            "bash": _bash_script,
            "zsh": _zsh_script,
            "tcsh": _tcsh_script,
        }
        print(generators[values](parser), end="")
        parser.exit()


def add_argument_to(parser: argparse.ArgumentParser, **_: object) -> None:
    parser.add_argument(
        "--print-completion",
        choices=("bash", "zsh", "tcsh"),
        action=_PrintCompletion,
        help="Print a shell completion script",
    )



from __future__ import annotations

import json
import os
import platform
import re
import shutil
import subprocess
import sys
from dataclasses import asdict, dataclass, field
from datetime import datetime, timezone
from io import StringIO
from typing import Any, Iterable, Sequence

import psutil


def _fmt_int(value: float | int | None, fallback: str = "??") -> str:
    if value is None:
        return fallback
    return str(int(value))


def _fmt_pct(value: float | int | None, fallback: str = "??") -> str:
    if value is None:
        return fallback
    return f"{float(value):.1f}%"


def _shorten(text: str, width: int) -> str:
    if width <= 0:
        return ""
    if len(text) <= width:
        return text.ljust(width)
    if width <= 3:
        return text[:width]
    return (text[: width - 3] + "...").ljust(width)


@dataclass
class TPUProcess:
    device_id: int
    pid: int | None
    process_name: str
    process_type: str = "C"
    memory_usage_mib: int = 0
    username: str | None = None
    command: str | None = None

    def display(
        self,
        *,
        show_user: bool,
        show_cmd: bool,
        show_pid: bool,
    ) -> str:
        parts: list[str] = []
        if show_user and self.username:
            parts.append(self.username)
        if show_cmd:
            parts.append(self.command or self.process_name or "?")
        base = ":".join(parts)
        if not base:
            base = self.process_name or "pid"
        if show_pid and self.pid is not None:
            base = f"{base}/{self.pid}"
        return f"{base}({self.memory_usage_mib}M)"


@dataclass
class TPUDevice:
    index: int
    name: str
    bus_id: str
    numa_node: int
    iommu_group: int
    power_draw_w: float
    power_cap_w: float
    hbm_used_mib: float
    hbm_total_mib: float
    duty_cycle_pct: float
    processes: list[TPUProcess] = field(default_factory=list)

    def print_to(
        self,
        fp: StringIO | Any,
        *,
        show_cmd: bool = False,
        show_user: bool = False,
        show_pid: bool = False,
        no_processes: bool = False,
        show_all: bool = False,
        tpuname_width: int = 16,
        color: bool = False,
        eol_char: str = os.linesep,
    ) -> Any:
        line = format_device_line(
            self,
            show_cmd=show_cmd,
            show_user=show_user,
            show_pid=show_pid,
            no_processes=no_processes,
            show_all=show_all,
            tpuname_width=tpuname_width,
            color=color,
        )
        fp.write(line)
        fp.write(eol_char)
        return fp


@dataclass
class TPUSnapshot:
    hostname: str
    query_time: datetime
    tool_version: str
    libtpu_version: str
    chip_type_name: str
    num_chips: int
    driver_version: str
    devices: list[TPUDevice]
    processes: list[TPUProcess]

    def to_dict(self) -> dict[str, Any]:
        payload = asdict(self)
        payload["query_time"] = self.query_time.isoformat()
        return payload


class Ansi:
    RESET = "\033[0m"
    CYAN = "\033[36m"
    BLUE = "\033[34m"
    GREEN = "\033[32m"
    BOLD_GREEN = "\033[1;32m"
    YELLOW = "\033[33m"
    BOLD_YELLOW = "\033[1;33m"
    MAGENTA = "\033[35m"
    GRAY = "\033[90m"
    BOLD = "\033[1m"


def _paint(text: str, code: str, enabled: bool) -> str:
    if not enabled:
        return text
    return f"{code}{text}{Ansi.RESET}"


def _processes_for_display(
    device: TPUDevice,
    *,
    show_cmd: bool,
    show_user: bool,
    show_pid: bool,
) -> str:
    if not device.processes:
        return "idle"
    return " ".join(
        process.display(show_cmd=show_cmd, show_user=show_user, show_pid=show_pid)
        for process in device.processes
    )


def format_device_line(
    device: TPUDevice,
    *,
    show_cmd: bool = False,
    show_user: bool = False,
    show_pid: bool = False,
    no_processes: bool = False,
    show_all: bool = False,
    tpuname_width: int = 16,
    color: bool = False,
) -> str:
    left = _paint(f"[{device.index}]", Ansi.CYAN, color)
    name = _paint(_shorten(device.name, tpuname_width), Ansi.BLUE, color)
    util = _paint(f"{_fmt_pct(device.duty_cycle_pct):>6}", Ansi.BOLD_GREEN, color)
    mem_used = _paint(f"{_fmt_int(device.hbm_used_mib):>5}", Ansi.BOLD_YELLOW, color)
    mem_total = _paint(f"{_fmt_int(device.hbm_total_mib):>5}", Ansi.YELLOW, color)

    sections = [f"{left} {name} | {util} | {mem_used} / {mem_total} MiB"]
    if show_all:
        sections.append(f"{device.bus_id} | NUMA {device.numa_node} | IOMMU {device.iommu_group}")
    if not no_processes:
        proc_text = _processes_for_display(
            device,
            show_cmd=show_cmd,
            show_user=show_user,
            show_pid=show_pid,
        )
        sections.append(_paint(proc_text, Ansi.GRAY, color))
    return " | ".join(sections)


class TPUStatCollection(Sequence[TPUDevice]):
    def __init__(self, snapshot: TPUSnapshot):
        self.snapshot = snapshot
        self.devices = snapshot.devices

    def __getitem__(self, index: int) -> TPUDevice:
        return self.devices[index]

    def __len__(self) -> int:
        return len(self.devices)

    @classmethod
    def new_query(cls, id: str | None = None, debug: bool = False) -> "TPUStatCollection":
        snapshot = query_snapshot(debug=debug)
        if id:
            wanted = {int(part.strip()) for part in id.split(",") if part.strip()}
            snapshot.devices = [device for device in snapshot.devices if device.index in wanted]
            snapshot.processes = [proc for proc in snapshot.processes if proc.device_id in wanted]
        return cls(snapshot)

    def print_json(self, fp: Any) -> None:
        json.dump(self.snapshot.to_dict(), fp, indent=2)
        fp.write(os.linesep)

    def print_formatted(
        self,
        fp: Any,
        *,
        show_cmd: bool = False,
        show_user: bool = False,
        show_pid: bool = False,
        no_processes: bool = False,
        show_all: bool = False,
        show_header: bool = True,
        force_color: bool = False,
        no_color: bool = False,
        tpuname_width: int = 16,
        color_stream: Any | None = None,
        eol_char: str = os.linesep,
    ) -> None:
        if color_stream is None:
            color_stream = fp
        color = should_use_color(force_color=force_color, no_color=no_color, stream=color_stream)
        if show_header:
            fp.write(format_header(self.snapshot, color=color))
            fp.write(eol_char)
        if not self.devices:
            fp.write("No TPUs detected.")
            fp.write(eol_char)
            return
        for device in self.devices:
            device.print_to(
                fp,
                show_cmd=show_cmd,
                show_user=show_user,
                show_pid=show_pid,
                no_processes=no_processes,
                show_all=show_all,
                tpuname_width=tpuname_width,
                color=color,
                eol_char=eol_char,
            )


def should_use_color(*, force_color: bool, no_color: bool, stream: Any) -> bool:
    if no_color:
        return False
    if force_color:
        return True
    if os.environ.get("NO_COLOR"):
        return False
    return bool(getattr(stream, "isatty", lambda: False)())


def format_header(snapshot: TPUSnapshot, *, color: bool = False) -> str:
    ts = snapshot.query_time.astimezone().strftime("%a %b %d %H:%M:%S %Y")
    host = _paint(snapshot.hostname, Ansi.BOLD, color)
    chip = _paint(f"[TPU {snapshot.chip_type_name} x{snapshot.num_chips}]", Ansi.CYAN, color)
    return f"{host}  {ts}  {chip}"


def query_snapshot(*, debug: bool = False) -> TPUSnapshot:
    raw = _collect_snapshot_raw()
    return _normalize_snapshot(raw, debug=debug)


def _collect_snapshot_raw() -> dict[str, Any]:
    raw = _collect_from_google_smi_api()
    if raw is not None:
        return raw
    raw = _collect_from_google_smi_cli()
    if raw is not None:
        return raw
    raise RuntimeError("Unable to collect TPU data from google_smi or google-smi.")


def _collect_from_google_smi_api() -> dict[str, Any] | None:
    try:
        from google_smi.collector import collect_snapshot
    except Exception:
        return None
    try:
        snapshot = collect_snapshot()
    except Exception:
        return None
    return asdict(snapshot)


def _collect_from_google_smi_cli() -> dict[str, Any] | None:
    if shutil.which("google-smi") is None:
        return None
    completed = subprocess.run(
        ["google-smi", "--json"],
        check=False,
        capture_output=True,
        text=True,
    )
    if completed.returncode != 0:
        raise RuntimeError(completed.stderr.strip() or "google-smi failed")
    return json.loads(completed.stdout)


def _normalize_snapshot(raw: dict[str, Any], *, debug: bool = False) -> TPUSnapshot:
    devices = []
    for entry in raw.get("devices", []):
        index = int(entry.get("device_id", 0))
        devices.append(
            TPUDevice(
                index=index,
                name=str(entry.get("chip_name", "TPU")),
                bus_id=str(entry.get("bus_id", "unknown")),
                numa_node=int(entry.get("numa_node", 0)),
                iommu_group=int(entry.get("iommu_group", 0)),
                power_draw_w=float(entry.get("power_draw_w", 0.0) or 0.0),
                power_cap_w=float(entry.get("power_cap_w", 0.0) or 0.0),
                hbm_used_mib=float(entry.get("hbm_used_mib", 0.0) or 0.0),
                hbm_total_mib=float(entry.get("hbm_total_mib", 0.0) or 0.0),
                duty_cycle_pct=float(entry.get("duty_cycle_pct", 0.0) or 0.0),
            )
        )

    process_map = _build_process_map(raw, devices, debug=debug)
    processes = [proc for device_processes in process_map.values() for proc in device_processes]
    for device in devices:
        device.processes = sorted(
            process_map.get(device.index, []),
            key=lambda proc: (
                proc.pid is None,
                proc.pid or 0,
                proc.process_name,
            ),
        )

    return TPUSnapshot(
        hostname=platform.node(),
        query_time=datetime.now(timezone.utc),
        tool_version=str(raw.get("tool_version", "unknown")),
        libtpu_version=str(raw.get("libtpu_version", "unknown")),
        chip_type_name=str(raw.get("chip_type_name", "unknown")),
        num_chips=int(raw.get("num_chips", len(devices))),
        driver_version=str(raw.get("driver_version", "unknown")),
        devices=devices,
        processes=processes,
    )


def _normalize_process(raw: dict[str, Any]) -> TPUProcess:
    pid = raw.get("pid")
    process = TPUProcess(
        device_id=int(raw.get("device_id", 0)),
        pid=int(pid) if pid is not None else None,
        process_name=str(raw.get("process_name", "?")),
        process_type=str(raw.get("process_type", "C")),
        memory_usage_mib=int(raw.get("memory_usage_mib", 0) or 0),
    )
    if process.pid is not None:
        _enrich_process(process)
    if not process.command:
        process.command = process.process_name
    return process


def _build_process_map(
    raw: dict[str, Any],
    devices: list[TPUDevice],
    *,
    debug: bool = False,
) -> dict[int, list[TPUProcess]]:
    process_map: dict[int, list[TPUProcess]] = {}
    seen: set[tuple[int, int | None, str]] = set()

    for entry in raw.get("processes", []):
        process = _normalize_process(entry)
        key = (process.device_id, process.pid, process.process_name)
        if key not in seen:
            process_map.setdefault(process.device_id, []).append(process)
            seen.add(key)

    try:
        owner_map = _scan_device_owners(devices, chip_type_name=str(raw.get("chip_type_name", "")))
    except Exception:
        if debug:
            raise
        owner_map = {}

    for device in devices:
        for pid in owner_map.get(device.index, []):
            process = TPUProcess(
                device_id=device.index,
                pid=pid,
                process_name="pid",
                memory_usage_mib=int(device.hbm_used_mib),
            )
            _enrich_process(process)
            if not process.command:
                process.command = process.process_name
            key = (process.device_id, process.pid, process.command or process.process_name)
            if key in seen:
                continue
            process_map.setdefault(process.device_id, []).append(process)
            seen.add(key)

    return process_map


def _scan_device_owners(devices: list[TPUDevice], *, chip_type_name: str) -> dict[int, list[int]]:
    path_to_index = _device_paths(devices, chip_type_name=chip_type_name)
    owners: dict[int, set[int]] = {device.index: set() for device in devices}
    pattern = re.compile(r"^/proc/(\d+)/fd/\d+$")
    for link in _iter_proc_fd_links():
        try:
            target = os.readlink(link)
        except FileNotFoundError:
            continue
        device_index = path_to_index.get(target)
        if device_index is None:
            continue
        match = pattern.fullmatch(link)
        if match:
            owners[device_index].add(int(match.group(1)))
    return {index: sorted(pids) for index, pids in owners.items() if pids}


def _iter_proc_fd_links() -> Iterable[str]:
    for proc_name in os.listdir("/proc"):
        if not proc_name.isdigit():
            continue
        fd_dir = os.path.join("/proc", proc_name, "fd")
        try:
            for fd_name in os.listdir(fd_dir):
                yield os.path.join(fd_dir, fd_name)
        except (FileNotFoundError, PermissionError, ProcessLookupError):
            continue


def _device_paths(devices: list[TPUDevice], *, chip_type_name: str) -> dict[str, int]:
    path_to_index: dict[str, int] = {}
    uses_vfio = chip_type_name.lower() in {"v5e", "v5p", "v6e", "v7x"}
    for device in devices:
        if uses_vfio:
            path_to_index[f"/dev/vfio/{device.index}"] = device.index
            path_to_index[f"/dev/vfio/{device.iommu_group}"] = device.index
        else:
            path_to_index[f"/dev/accel{device.index}"] = device.index
    return path_to_index


def _enrich_process(process: TPUProcess) -> None:
    try:
        ps_process = psutil.Process(process.pid)
        process.username = ps_process.username()
        cmdline = ps_process.cmdline()
        if cmdline:
            process.command = os.path.basename(cmdline[0])
    except (psutil.Error, FileNotFoundError, PermissionError):
        return


def clear_screen(stream: Any) -> None:
    stream.write("\033[H\033[J")
    stream.flush()


def move_cursor_home(stream: Any) -> None:
    stream.write("\033[H")


def clear_to_end(stream: Any) -> None:
    stream.write("\033[J")


def hide_cursor(stream: Any) -> None:
    stream.write("\033[?25l")


def show_cursor(stream: Any) -> None:
    stream.write("\033[?25h")



from __future__ import annotations

import argparse
import signal
import sys
import time
from io import StringIO

try:
    import shtab
except ImportError:
    from tpustat import _shtab as shtab

from tpustat import __version__
from tpustat.core import (
    TPUStatCollection,
    clear_screen,
    clear_to_end,
    hide_cursor,
    show_cursor,
)


def print_tpustat(*, json: bool = False, **kwargs: object) -> None:
    stream = kwargs.pop("stream", sys.stdout)
    color_stream = kwargs.pop("color_stream", stream)
    debug = bool(kwargs.pop("debug", False))
    stats = TPUStatCollection.new_query(id=kwargs.pop("id", None), debug=debug)
    if json:
        stats.print_json(stream)
    else:
        stats.print_formatted(stream, color_stream=color_stream, **kwargs)


def _render_frame(**kwargs: object) -> str:
    buf = StringIO()
    print_tpustat(stream=buf, color_stream=sys.stdout, **kwargs)
    return buf.getvalue()


def _write_tty_frame(stream: object, frame: str, previous_line_count: int) -> int:
    lines = frame.splitlines()
    stream.write("\033[H")
    for line in lines:
        stream.write("\033[2K")
        stream.write(line)
        stream.write("\n")
    for _ in range(max(0, previous_line_count - len(lines))):
        stream.write("\033[2K\n")
    clear_to_end(stream)
    return len(lines)


def loop_tpustat(interval: float, **kwargs: object) -> int:
    is_tty = bool(getattr(sys.stdout, "isatty", lambda: False)())
    previous_line_count = 0
    try:
        if is_tty:
            clear_screen(sys.stdout)
            hide_cursor(sys.stdout)
        while True:
            if is_tty:
                frame = _render_frame(**kwargs)
                previous_line_count = _write_tty_frame(sys.stdout, frame, previous_line_count)
            else:
                print_tpustat(**kwargs)
            sys.stdout.flush()
            time.sleep(interval)
    except KeyboardInterrupt:
        return 0
    finally:
        if is_tty:
            show_cursor(sys.stdout)
            sys.stdout.flush()


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser("tpustat")
    shtab.add_argument_to(parser)
    parser_color = parser.add_mutually_exclusive_group()
    parser_color.add_argument("--force-color", "--color", action="store_true", help="Force colored output")
    parser_color.add_argument("--no-color", action="store_true", help="Disable colored output")
    parser.add_argument("--id", help="Target a specific TPU index or comma-separated indices")
    parser.add_argument("-a", "--show-all", action="store_true", help="Show bus, NUMA, and IOMMU details")
    parser.add_argument("-c", "--show-cmd", action="store_true", help="Display process command names")
    parser.add_argument("-u", "--show-user", action="store_true", help="Display process owners")
    parser.add_argument("-p", "--show-pid", action="store_true", help="Display process IDs")
    parser.add_argument("--tpuname-width", type=nonnegative_int, default=16, help="Width used for TPU names")
    parser.add_argument("--json", action="store_true", help="Print snapshot data as JSON")
    parser.add_argument("-i", "--interval", "--watch", nargs="?", type=float, default=0, help="Refresh continuously")
    parser.add_argument("--no-header", dest="show_header", action="store_false", default=True, help="Suppress the header")
    parser.add_argument("--no-processes", action="store_true", help="Do not display process information")
    parser.add_argument("--debug", action="store_true", help="Show full errors when TPU data collection fails")
    parser.add_argument("-v", "--version", action="version", version=f"tpustat {__version__}")
    return parser


def nonnegative_int(value: str) -> int:
    parsed = int(value)
    if parsed < 0:
        raise argparse.ArgumentTypeError("Only non-negative integers are allowed.")
    return parsed


def main(argv: list[str] | None = None) -> int:
    if argv is None:
        argv = sys.argv[1:]

    try:
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    except Exception:
        pass

    parser = build_parser()
    args = parser.parse_args(argv)
    if hasattr(args, "print_completion"):
        delattr(args, "print_completion")
    if args.show_all:
        args.show_cmd = True
        args.show_user = True
        args.show_pid = True

    if args.interval is None:
        args.interval = 1.0
    if args.interval and args.interval > 0:
        if args.json:
            parser.error("--json and --interval/-i cannot be used together")
        interval = max(0.1, args.interval)
        kwargs = vars(args)
        kwargs.pop("interval", None)
        return loop_tpustat(interval=interval, **kwargs)

    kwargs = vars(args)
    kwargs.pop("interval", None)
    try:
        print_tpustat(**kwargs)
    except Exception as exc:
        sys.stderr.write("Error while querying TPU devices. Use --debug for details.\n")
        sys.stderr.write(f"{exc}\n")
        if args.debug:
            raise
        return 1
    return 0



from __future__ import annotations

from io import StringIO
import os

from tpustat.core import (
    TPUStatCollection,
    _device_paths,
    _normalize_snapshot,
    _scan_device_owners,
    format_device_line,
    format_header,
)


def _no_device_owners(devices, chip_type_name):
    del devices, chip_type_name
    return {}


def test_normalize_snapshot_attaches_processes(monkeypatch, raw_snapshot):
    monkeypatch.setattr("tpustat.core._scan_device_owners", _no_device_owners)
    snapshot = _normalize_snapshot(raw_snapshot, debug=False)

    assert snapshot.num_chips == 2
    assert len(snapshot.devices) == 2
    assert snapshot.devices[0].processes[0].pid == 1234
    assert snapshot.devices[1].processes == []


def test_format_device_line_default(monkeypatch, raw_snapshot):
    monkeypatch.setattr("tpustat.core._scan_device_owners", _no_device_owners)
    snapshot = _normalize_snapshot(raw_snapshot, debug=False)

    line = format_device_line(snapshot.devices[0], show_cmd=True, show_pid=True)

    assert "[0]" in line
    assert "TPU v6e" in line
    assert "42.5%" in line
    assert "512 / 32768 MiB" in line
    assert "python/1234(512M)" in line


def test_format_device_line_show_all(monkeypatch, raw_snapshot):
    monkeypatch.setattr("tpustat.core._scan_device_owners", _no_device_owners)
    snapshot = _normalize_snapshot(raw_snapshot, debug=False)

    line = format_device_line(snapshot.devices[0], show_all=True)

    assert "0000:00:04.0" in line
    assert "NUMA 0" in line
    assert "IOMMU 0" in line
    assert "PCIe" not in line


def test_collection_print_formatted(monkeypatch, raw_snapshot):
    monkeypatch.setattr("tpustat.core._scan_device_owners", _no_device_owners)
    snapshot = _normalize_snapshot(raw_snapshot, debug=False)
    stats = TPUStatCollection(snapshot)
    buf = StringIO()

    stats.print_formatted(buf, show_cmd=True, show_pid=True, force_color=False, no_color=True)
    output = buf.getvalue()

    assert "TPU v6e x2" in output
    assert "python/1234(512M)" in output
    assert "idle" in output


def test_format_header_contains_chip_summary(monkeypatch, raw_snapshot):
    monkeypatch.setattr("tpustat.core._scan_device_owners", _no_device_owners)
    snapshot = _normalize_snapshot(raw_snapshot, debug=False)

    header = format_header(snapshot)

    assert "[TPU v6e x2]" in header
    assert "vfio-pci" not in header


def test_format_device_line_respects_name_width(monkeypatch, raw_snapshot):
    monkeypatch.setattr("tpustat.core._scan_device_owners", _no_device_owners)
    snapshot = _normalize_snapshot(raw_snapshot, debug=False)

    line = format_device_line(snapshot.devices[0], tpuname_width=6)

    assert "TPU..." in line


def test_device_paths_include_vfio_and_iommu(monkeypatch, raw_snapshot):
    monkeypatch.setattr("tpustat.core._scan_device_owners", _no_device_owners)
    snapshot = _normalize_snapshot(raw_snapshot, debug=False)

    path_map = _device_paths(snapshot.devices, chip_type_name=snapshot.chip_type_name)

    assert path_map["/dev/vfio/0"] == 0
    assert path_map["/dev/vfio/1"] == 1


def test_scan_device_owners(monkeypatch, raw_snapshot):
    monkeypatch.setattr("tpustat.core._scan_device_owners", _no_device_owners)
    snapshot = _normalize_snapshot(raw_snapshot)

    monkeypatch.setattr("tpustat.core._iter_proc_fd_links", lambda: ["/proc/2000/fd/7", "/proc/2001/fd/8"])

    def fake_readlink(path: str) -> str:
        return {
            "/proc/2000/fd/7": "/dev/vfio/0",
            "/proc/2001/fd/8": "/dev/vfio/1",
        }[path]

    monkeypatch.setattr(os, "readlink", fake_readlink)

    owners = _scan_device_owners(snapshot.devices, chip_type_name=snapshot.chip_type_name)

    assert owners == {0: [2000], 1: [2001]}



from __future__ import annotations

import json
from io import StringIO

import pytest

from tpustat import cli
from tpustat.core import TPUStatCollection, _normalize_snapshot


@pytest.fixture
def stats(monkeypatch, raw_snapshot):
    monkeypatch.setattr("tpustat.core._scan_device_owners", lambda devices, chip_type_name: {})
    return TPUStatCollection(_normalize_snapshot(raw_snapshot))


def test_cli_json_output(monkeypatch, capsys, stats):
    monkeypatch.setattr(cli.TPUStatCollection, "new_query", classmethod(lambda cls, id=None, debug=False: stats))

    exit_code = cli.main(["--json"])
    captured = capsys.readouterr()

    assert exit_code == 0
    payload = json.loads(captured.out)
    assert payload["chip_type_name"] == "v6e"
    assert payload["devices"][0]["index"] == 0
    assert "pcie_gen" not in payload["devices"][0]
    assert "pcie_width" not in payload["devices"][0]


def test_cli_text_output(monkeypatch, capsys, stats):
    monkeypatch.setattr(cli.TPUStatCollection, "new_query", classmethod(lambda cls, id=None, debug=False: stats))

    exit_code = cli.main(["-c", "-p", "--no-color"])
    captured = capsys.readouterr()

    assert exit_code == 0
    assert "python/1234(512M)" in captured.out


def test_cli_show_all_expands_flags(monkeypatch, capsys, stats):
    monkeypatch.setattr(cli.TPUStatCollection, "new_query", classmethod(lambda cls, id=None, debug=False: stats))

    exit_code = cli.main(["--show-all", "--no-color"])
    captured = capsys.readouterr()

    assert exit_code == 0
    assert "0000:00:04.0" in captured.out
    assert "/1234" in captured.out


def test_cli_reports_query_errors(monkeypatch, capsys):
    def blow_up(cls, id=None, debug=False):
        raise RuntimeError("collector exploded")

    monkeypatch.setattr(cli.TPUStatCollection, "new_query", classmethod(blow_up))

    exit_code = cli.main(["--no-color"])
    captured = capsys.readouterr()

    assert exit_code == 1
    assert "collector exploded" in captured.err


def test_cli_rejects_json_watch(monkeypatch):
    monkeypatch.setattr(
        cli.TPUStatCollection,
        "new_query",
        classmethod(lambda cls, id=None, debug=False: TPUStatCollection(_normalize_snapshot({
            "tool_version": "0.1.0",
            "libtpu_version": "0.0.17",
            "chip_type_name": "v6e",
            "num_chips": 0,
            "driver_version": "vfio-pci",
            "devices": [],
            "processes": [],
        }))),
    )

    with pytest.raises(SystemExit):
        cli.main(["--json", "-i", "1"])


def test_parser_has_completion_flag(capsys):
    parser = cli.build_parser()
    with pytest.raises(SystemExit) as exc:
        parser.parse_args(["--print-completion", "bash"])
    captured = capsys.readouterr()

    assert exc.value.code == 0
    assert "tpustat" in captured.out
    assert "complete " in captured.out or "#compdef tpustat" in captured.out


def test_write_tty_frame_clears_stale_line_content():
    buf = StringIO()

    line_count = cli._write_tty_frame(
        buf,
        "header\n[0] TPU v6e | 5758 / 31995 MiB | pid(5758M) train.py(5758M)\n",
        previous_line_count=0,
    )
    line_count = cli._write_tty_frame(
        buf,
        "header\n[0] TPU v6e |    0 / 32768 MiB | idle\n",
        previous_line_count=line_count,
    )
    output = buf.getvalue()

    assert line_count == 2
    assert "\033[2K[0] TPU v6e |    0 / 32768 MiB | idle\n" in output
    assert "idle5758M) train.py(5758M)" not in output


def test_print_tpustat_uses_color_stream_for_buffered_output(monkeypatch, stats):
    class FakeTTY(StringIO):
        def isatty(self) -> bool:
            return True

    monkeypatch.delenv("NO_COLOR", raising=False)
    monkeypatch.setattr(cli.TPUStatCollection, "new_query", classmethod(lambda cls, id=None, debug=False: stats))

    buf = StringIO()
    cli.print_tpustat(stream=buf, color_stream=FakeTTY())

    assert "\033[" in buf.getvalue()



from pathlib import Path

from setuptools import setup


ROOT = Path(__file__).parent


setup(
    name="tpustat",
    version="0.1.1",
    description="A gpustat-style TPU status CLI built on top of google-smi.",
    long_description=(ROOT / "README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    author="Minho Ryu",
    license="MIT",
    python_requires=">=3.10",
    packages=["tpustat"],
    include_package_data=True,
    install_requires=[
        "google-smi>=0.1.0",
        "psutil>=5.9",
        "shtab>=1.7",
    ],
    extras_require={
        "dev": ["pytest>=8.0"],
    },
    entry_points={
        "console_scripts": [
            "tpustat=tpustat.cli:main",
        ],
    },
)
